package com.kuaishou.common.netty;

import java.util.Arrays;
import java.util.List;

import com.kuaishou.common.netty.annotation.NotSharable;

import io.netty.buffer.ByteBuf;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.ByteToMessageDecoder;
import io.netty.handler.codec.CorruptedFrameException;

/**
 * +----------------------------------+
 * |version(8bit)|magic number(24bit) |
 * +----------------------------------+
 * |    header reserved(64bit)        |
 * +----------------------------------+
 * |      body length (32bit)         |
 * +----------------------------------+
 * |      Body (Protobuf)             |
 * +----------------------------------+
 *
 * @author Li Weibo  //I believe in Spring Brother.
 *
 */
@NotSharable
public abstract class AbstractSocketMessageDecoder extends ByteToMessageDecoder {

    @Override
    protected void decode(ChannelHandlerContext ctx, ByteBuf in, List<Object> out)
            throws Exception {

        if (in.readableBytes() < Constants.HEADER_LENGTH) {
            return;
        }

        //标记读取位置，以便下面in.resetReaderIndex()的时候
        //能够回到此位置，否则可能会出现死循in.markReaderIndex();
        in.markReaderIndex();
        byte version = in.readByte();
        if (version != Constants.PROTOCOL_VERSION) {
            in.clear(); //格式已坏, 清数据, 否则channel关闭之前此方法还会被进入几次
            throw new CorruptedFrameException("bad version: " + version);
        }
        byte[] magicBytes = new byte[Constants.MAGIC_BYTES.length];
        in.readBytes(magicBytes);
        if (!Arrays.equals(Constants.MAGIC_BYTES, magicBytes)) {
            in.clear();
            throw new CorruptedFrameException("Bad magic bytes:" + Arrays.toString(magicBytes));
        }
        in.readBytes(new byte[Constants.HEADER_RESERVED_BYTES]); //header reserved(64bit)
        int length = in.readInt();
        if (length <= 0 || length > Constants.MAX_BODY_LENGTH) {
            in.clear();
            throw new CorruptedFrameException("Bad length:" + length);
        }
        if (in.readableBytes() < length) {
            in.resetReaderIndex();
            return;
        }
        //解包放入pipeline
        byte[] data = new byte[length];
        in.readBytes(data);
        out.add(parseFrom(data));
    }

    public abstract Object parseFrom(byte[] data) throws Exception;
}
